from typing import Sequence, Callable, List
import numpy as np
import jax
import jax.numpy as jnp
import distrax

from flax import linen as nn  # The Linen API


class MLP(nn.Module):
    features: Sequence[int]
    activation: Callable = nn.tanh
    use_bias: bool = True
    init_weight_scale: float = 1e-2
    kernel_i: Callable = jax.nn.initializers.variance_scaling

    def setup(self):
        self.layers = [
            nn.Dense(
                feat,
                use_bias=self.use_bias,
                kernel_init=self.kernel_i(self.init_weight_scale, "fan_in", "normal"),
            )
            for feat in self.features
        ]

    def __call__(self, x):
        for l, layer in enumerate(self.layers[:-1]):
            x = self.activation(layer(x))
        x = self.layers[-1](x)
        return x


class Reshape(nn.Module):
    shape: Sequence[int]

    def __call__(self, x):
        return jnp.reshape(x.T, self.shape)


class Conditioner(nn.Module):
    n_features: int
    hidden_size: Sequence[int]
    num_bijector_params: int

    def setup(self):
        self.conditioner = nn.Sequential(
            [
                MLP([self.n_features] + list(self.hidden_size)),
                nn.tanh,
                nn.Dense(
                    self.n_features * self.num_bijector_params,
                    kernel_init=jax.nn.initializers.zeros,
                    bias_init=jax.nn.initializers.zeros,
                ),
                Reshape((self.n_features, self.num_bijector_params)),
            ]
        )

    def __call__(self, x):
        return self.conditioner(x)


class Scalar(nn.Module):
    n_features: int

    def setup(self):
        self.shift = self.param(
            "shifts", lambda rng, shape: jnp.zeros(shape), (self.n_features)
        )
        self.scale = self.param(
            "scales", lambda rng, shape: jnp.ones(shape), (self.n_features)
        )

    def __call__(self, x):
        return self.scale, self.shift


def scalar_affine(params: jnp.ndarray):
    return distrax.ScalarAffine(scale=params[0], shift=params[1])


class RQSpline(nn.Module):

    """
    Rational quadratic spline normalizing flow model using distrax.

    Parameters
    ----------
    n_features : int
        Number of features in the data.
    num_layers : int
        Number of layers in the flow.
    num_bins : int
        Number of bins in the spline.
    hidden_size : Sequence[int]
        Size of the hidden layers in the conditioner.
    spline_range : Sequence[float]
        Range of the spline.
    """

    n_features: int
    num_layers: int
    hidden_size: Sequence[int]
    num_bins: int
    spline_range: Sequence[float] = (-10.0, 10.0)

    def setup(self):
        conditioner = []
        scalar = []
        for i in range(self.num_layers):
            conditioner.append(
                Conditioner(self.n_features, self.hidden_size, 3 * self.num_bins + 1)
            )
            scalar.append(Scalar(self.n_features))

        self.conditioner = conditioner
        self.scalar = scalar

        self.base_mean = self.variable(
            "variables", "base_mean", jnp.zeros, ((self.n_features))
        )
        self.base_cov = self.variable(
            "variables", "base_cov", jnp.eye, (self.n_features)
        )

        self.vmap_call = jax.jit(jax.vmap(self.__call__))

        def bijector_fn(params: jnp.ndarray):
            return distrax.RationalQuadraticSpline(
                params, range_min=self.spline_range[0], range_max=self.spline_range[1]
            )

        self.bijector_fn = bijector_fn

    def make_flow(self):
        mask = (jnp.arange(0, self.n_features) % 2).astype(bool)
        mask_all = (jnp.zeros(self.n_features)).astype(bool)
        layers = []
        for i in range(self.num_layers):
            layers.append(
                distrax.MaskedCoupling(
                    mask=mask_all, bijector=scalar_affine, conditioner=self.scalar[i]
                )
            )
            layers.append(
                distrax.MaskedCoupling(
                    mask=mask,
                    bijector=self.bijector_fn,
                    conditioner=self.conditioner[i],
                )
            )
            mask = jnp.logical_not(mask)

        flow = distrax.Inverse(distrax.Chain(layers))
        base_dist = distrax.Independent(
            distrax.MultivariateNormalFullCovariance(
                loc=jnp.zeros(self.n_features),
                covariance_matrix=jnp.eye(self.n_features),
            )
        )

        return base_dist, flow

    def __call__(self, x: jnp.array) -> jnp.array:
        x = (x-self.base_mean.value)/jnp.sqrt(jnp.diag(self.base_cov.value))
        base_dist, flow = self.make_flow()
        return distrax.Transformed(base_dist, flow).log_prob(x)

    def sample(self, rng: jax.random.PRNGKey, num_samples: int) -> jnp.array:
        base_dist, flow = self.make_flow()
        samples = distrax.Transformed(base_dist, flow).sample(
            seed=rng, sample_shape=(num_samples)
        )
        return samples * jnp.sqrt(jnp.diag(self.base_cov.value)) + self.base_mean.value

    def log_prob(self, x: jnp.array) -> jnp.array:
        return self.vmap_call(x)



import jax
import jax.numpy as jnp
from flowMC.nfmodel.rqSpline import RQSpline
from flowMC.sampler.MALA import make_mala_sampler
from flowMC.sampler.Sampler import Sampler
from flowMC.utils.PRNG_keys import initialize_rng_keys
from flowMC.nfmodel.utils import *

def log_prob(x):
    return -0.5 * jnp.sum(x ** 2)

n_dim = 5
n_chains = 10

rng_key_set = initialize_rng_keys(n_chains, seed=42)
initial_position = jax.random.normal(rng_key_set[0], shape=(n_chains, n_dim)) * 1
model = RQSpline(n_dim, 3, [64, 64], 8)
local_sampler_caller = lambda x: make_mala_sampler(x, jit=True)
sampler_params = {'dt': 1e-1}

nf_sampler = Sampler(n_dim, rng_key_set, local_sampler_caller, sampler_params, log_prob,
                    model,
                    n_chains=n_chains)

nf_sampler.sample(initial_position)




from typing import Sequence, Callable
import jax
import jax.numpy as jnp
from flax import linen as nn
import numpy as np


class MLP(nn.Module):

    """
    Multi-layer perceptron in Flax.

    Parameters
    ----------
    features : list of int
        The number of features in each layer.
    activation : callable
        The activation function at each level
    use_bias : bool
        Whether to use bias in the layers.
    init_weight_scale : float
        The initial weight scale for the layers.
    kernel_init : callable
        The kernel initializer for the layers.
        We use a gaussian kernel with a standard deviation of `init_weight_scale` by default.
    """

    features: Sequence[int]
    activation: Callable = nn.relu
    use_bias: bool = True
    init_weight_scale: float = 1e-4
    kernel_i: Callable = jax.nn.initializers.variance_scaling

    def setup(self):
        self.layers = [
            nn.Dense(
                feat,
                use_bias=self.use_bias,
                kernel_init=self.kernel_i(self.init_weight_scale, "fan_in", "normal"),
            )
            for feat in self.features
        ]

    def __call__(self, x):
        for l, layer in enumerate(self.layers[:-1]):
            x = self.activation(layer(x))
        x = self.layers[-1](x)
        return x


class AffineCoupling(nn.Module):

    """
    Affine coupling layer. (Defined in the RealNVP paper https://arxiv.org/abs/1605.08803)
    We use tanh as the default activation function.

    Parameters
    ----------
    n_features : int
        The number of features in the input.
    n_hidden : int
        The number of hidden units in the MLP.
    mask : ndarray
        Alternating mask for the affine coupling layer.
    dt : float
        Scaling factor for the affine coupling layer.
    """

    n_features: int
    n_hidden: int
    mask: jnp.array
    dt: float = 1

    def setup(self):
        self.scale_MLP = MLP([self.n_features, self.n_hidden, self.n_features])
        self.translate_MLP = MLP([self.n_features, self.n_hidden, self.n_features])

    def __call__(self, x):
        s = self.mask * self.scale_MLP(x * (1 - self.mask))
        s = jnp.tanh(s)
        t = self.mask * self.translate_MLP(x * (1 - self.mask))
        s = self.dt * s
        t = self.dt * t
        log_det = s.reshape(s.shape[0], -1).sum(axis=-1)
        outputs = (x + t) * jnp.exp(s)
        return outputs, log_det

    def inverse(self, x):
        s = self.mask * self.scale_MLP(x * (1 - self.mask))
        s = jnp.tanh(s)
        t = self.mask * self.translate_MLP(x * (1 - self.mask))
        s = self.dt * s
        t = self.dt * t
        log_det = -s.reshape(s.shape[0], -1).sum(axis=-1)
        outputs = x * jnp.exp(-s) - t
        return outputs, log_det


class RealNVP(nn.Module):

    """

    RealNVP mode defined in the paper https://arxiv.org/abs/1605.08803.
    MLP is needed to make sure the scaling between layers are more or less the same.

    Parameters
    ----------
    n_features : int
        The number of features in the input.
    n_hidden : int
        The number of hidden units in the MLP.
    n_layer : int
        The number of affine coupling layers.
    dt : float
        Scaling factor for the affine coupling layer.
    base_mean : ndarray
        Mean of Gaussian base distribution
    base_cov : ndarray
        Covariance of Gaussian base distribution
    """

    n_layer: int
    n_features: int
    n_hidden: int
    dt: float = 1

    def setup(self):
        affine_coupling = []
        for i in range(self.n_layer):
            mask = np.ones(self.n_features)
            mask[int(self.n_features / 2) :] = 0
            if i % 2 == 0:
                mask = 1 - mask
            mask = jnp.array(mask)
            affine_coupling.append(
                AffineCoupling(self.n_features, self.n_hidden, mask, dt=self.dt)
            )
        self.affine_coupling = affine_coupling

        self.base_mean = self.variable(
            "variables", "base_mean", jnp.zeros, ((self.n_features))
        )
        self.base_cov = self.variable(
            "variables", "base_cov", jnp.eye, (self.n_features)
        )

    def __call__(self, x):
        log_det = jnp.zeros(x.shape[0])
        for i in range(self.n_layer):
            x, log_det_i = self.affine_coupling[i](x)
            log_det += log_det_i
        return x, log_det

    def inverse(self, x):
        x = (x-self.base_mean.value)/jnp.sqrt(jnp.diag(self.base_cov.value))
        log_det = jnp.zeros(x.shape[0])
        for i in range(self.n_layer):
            x, log_det_i = self.affine_coupling[self.n_layer - 1 - i].inverse(x)
            log_det += log_det_i
        return x, log_det

    def sample(self, rng_key, n_samples):
        gaussian = jax.random.multivariate_normal(
            rng_key, jnp.zeros(self.n_features), jnp.eye(self.n_features), shape=(n_samples,)
        )
        samples = self.inverse(gaussian)[0]
        samples = samples * jnp.sqrt(jnp.diag(self.base_cov.value)) + self.base_mean.value
        return samples # Return only the samples 

    def log_prob(self, x):
        x = (x-self.base_mean.value)/jnp.sqrt(jnp.diag(self.base_cov.value))
        y, log_det = self.__call__(x)
        log_det = log_det + jax.scipy.stats.multivariate_normal.logpdf(
            y, jnp.zeros(self.n_features), jnp.eye(self.n_features)
        )
        return log_det



import jax
import jax.numpy as jnp
from jax.scipy.stats import multivariate_normal
from tqdm import tqdm


def make_mMALA_kernel(logpdf, d_logpdf, d2_logpdf, dt):
    def mMALA_kernel(rng_key, position, log_prob):

        """
        Metropolis-adjusted Langevin algorithm kernel.
        This function make a proposal and accept/reject it.

        Args:
            rng_key (n_chains, 2): random key
            logpdf (function) : log-density function
            d_logpdf (function): gradient of log-density function
            position (n_chains, n_dim): current position
            log_prob (n_chains, ): log-probability of the current position
            dt (float): step size of the MALA step

        Returns:
            position (n_chains, n_dim): the new poisiton of the chain
            log_prob (n_chains, ): the log-probability of the new position
            do_accept (n_chains, ): whether to accept the new position

        """
        key1, key2 = jax.random.split(rng_key)

        n_dim = position.shape[-1]
        d_log_current = d_logpdf(position)
        d2_current = d2_logpdf(position)
        d2_current = jax.lax.cond(
            jnp.any(jnp.isnan(d2_current)), lambda: jnp.eye(n_dim), lambda: d2_current
        )
        d2_current = jnp.linalg.inv(d2_current)

        dt_current_sqrt = dt * jnp.sqrt(d2_current)

        proposal = position + jnp.dot((dt * d2_current) ** 2, d_log_current) / 2
        proposal += jnp.dot(
            dt_current_sqrt, jax.random.normal(key1, shape=position.shape)
        )
        proposal_log_prob = logpdf(proposal)
        d2_proposal = d2_logpdf(proposal)
        d2_proposal = jax.lax.cond(
            jnp.any(jnp.isnan(d2_proposal)), lambda: jnp.eye(n_dim), lambda: d2_proposal
        )
        d2_proposal = jnp.linalg.inv(d2_proposal)

        ratio = proposal_log_prob - logpdf(position)
        ratio -= multivariate_normal.logpdf(
            position,
            proposal + jnp.dot(d2_proposal * dt * dt, d_logpdf(proposal)) / 2,
            d2_proposal * dt * dt,
        )
        ratio += multivariate_normal.logpdf(
            proposal,
            position + jnp.dot(d2_current * dt * dt, d_log_current) / 2,
            d2_current * dt * dt,
        )

        log_uniform = jnp.log(jax.random.uniform(key2))
        do_accept = log_uniform < ratio

        position = jnp.where(do_accept, proposal, position)
        log_prob = jnp.where(do_accept, proposal_log_prob, log_prob)
        return position, log_prob, do_accept

    return mMALA_kernel


def make_mMALA_update(logpdf, d_logpdf, d2_logpdf, dt):
    mMALA_kernel = make_mMALA_kernel(logpdf, d_logpdf, d2_logpdf, dt)

    def mMALA_update(i, state):
        key, positions, log_prob, acceptance = state
        _, key = jax.random.split(key)
        new_position, new_log_prob, do_accept = mMALA_kernel(
            key, positions[i - 1], log_prob[i - 1]
        )
        positions = positions.at[i].set(new_position)
        log_prob = log_prob.at[i].set(new_log_prob)
        acceptance = acceptance.at[i].set(do_accept)
        return (key, positions, log_prob, acceptance)

    mMALA_update = jax.vmap(mMALA_update, in_axes=(None, (0, 0, 0, 0)))
    mMALA_kernel_vec = jax.vmap(mMALA_kernel, in_axes=(0, 0, 0))
    logpdf = jax.vmap(logpdf)
    d_logpdf = jax.vmap(d_logpdf)

    # Apperantly jitting after vmap will make compilation much slower.
    # Output the kernel, logpdf, and dlogpdf for warmup jitting.
    # Apperantly passing in a warmed up function will still trigger recompilation.
    # so the warmup need to be done with the output function

    return mMALA_update, mMALA_kernel_vec, logpdf, d_logpdf


def make_MALA_sampler(logpdf, dt=1e-5, jit=False):

    d_logpdf = jax.grad(logpdf)
    d2_logpdf = jax.hessian(logpdf)
    mMALA_update, mk, lp, dlp = make_mMALA_update(logpdf, d_logpdf, d2_logpdf, dt)
    # Somehow if I define the function inside the other function,
    # I think it doesn't use the cache and recompile everytime.
    if jit:
        mMALA_update = jax.jit(mMALA_update)
        mk = jax.jit(mk)
        lp = jax.jit(lp)
        dlp = jax.jit(dlp)
        d2lp = jax.jit(d2lp)

    def mMALA_sampler(rng_key, n_steps, initial_position):

        """
        Metropolis-adjusted Langevin algorithm sampler.
        This function do n step with the mMALA kernel.

        Args:
            rng_key (n_chains, 2): random key for the sampler
            n_steps (int): number of local steps
            logpdf (function): log-density function
            d_logpdf (function): gradient of log-density function
            initial_position (n_chains, n_dim): initial position of the chain
            dt (float): step size of the mMALA step

        Returns:
            rng_key (n_chains, 2): random key for the sampler after the sampling
            all_positions (n_chains, n_steps, n_dim): all the positions of the chain
            log_probs (n_chains, ): log probability at the end of the chain
            acceptance: acceptance rate of the chain
        """

        logp = lp(initial_position)
        n_chains = rng_key.shape[0]
        acceptance = jnp.zeros(
            (
                n_chains,
                n_steps,
            )
        )
        all_positions = (
            jnp.zeros(
                (
                    n_chains,
                    n_steps,
                )
                + initial_position.shape[-1:]
            )
            + initial_position[:, None]
        )
        all_logp = (
            jnp.zeros(
                (
                    n_chains,
                    n_steps,
                )
            )
            + logp[:, None]
        )
        state = (rng_key, all_positions, all_logp, acceptance)
        for i in tqdm(
            range(1, n_steps), desc="Sampling Locally", miniters=int(n_steps / 10)
        ):
            state = mMALA_update(i, state)
        return state

    return mMALA_sampler, mMALA_update, mk, lp, dlp



import jax
import jax.numpy as jnp  # JAX NumPy
import jax.random as random  # JAX random
from tqdm import trange


def make_training_loop(model):
    """
    Create a function that trains an NF model.

    Args:
        model: a neural network model with a `loss` method.

    Returns:
        train_flow (Callable): wrapper function that trains the model.

    """

    def train_step(batch, state, variables):
        def loss(params):
            log_det = model.apply(
                {"params": params, "variables": variables}, batch, method=model.log_prob
            )
            return -jnp.mean(log_det)

        grad_fn = jax.value_and_grad(loss)
        value, grad = grad_fn(state.params)
        state = state.apply_gradients(grads=grad)
        return value, state

    train_step = jax.jit(train_step)

    def train_epoch(rng, state, variables, train_ds, batch_size):
        """Train for a single epoch."""
        train_ds_size = len(train_ds)
        steps_per_epoch = train_ds_size // batch_size
        if steps_per_epoch > 0:
            perms = jax.random.permutation(rng, train_ds_size)

            perms = perms[: steps_per_epoch * batch_size]  # skip incomplete batch
            perms = perms.reshape((steps_per_epoch, batch_size))
            for perm in perms:
                batch = train_ds[perm, ...]
                value, state = train_step(batch, state, variables)
        else:
            value, state = train_step(train_ds, state, variables)

        return value, state

    def train_flow(rng, state, variables, data, num_epochs, batch_size):
        loss_values = jnp.zeros(num_epochs)
        pbar = trange(num_epochs, desc="Training NF", miniters=int(num_epochs / 10))
        best_state = state
        best_loss = 1e9
        for epoch in pbar:
            # Use a separate PRNG key to permute image data during shuffling
            rng, input_rng = jax.random.split(rng)
            # Run an optimization step over a training batch
            value, state = train_epoch(input_rng, state, variables, data, batch_size)
            # print('Train loss: %.3f' % value)
            loss_values = loss_values.at[epoch].set(value)
            if loss_values[epoch] < best_loss:
                best_state = state
                best_loss = loss_values[epoch]
            if num_epochs > 10:
                if epoch % int(num_epochs / 10) == 0:
                    pbar.set_description(f"Training NF, current loss: {value:.3f}")
            else:
                if epoch == num_epochs:
                    pbar.set_description(f"Training NF, current loss: {value:.3f}")

        return rng, best_state, loss_values

    return train_flow, train_epoch, train_step


def sample_nf(model, param, rng_key, n_sample, variables):
    rng_key, subkey = random.split(rng_key)
    samples = model.apply(
        {"params": param, "variables": variables}, subkey, n_sample, method=model.sample
    )
    # samples = jnp.flip(samples[0],axis=1)
    return rng_key, samples



from tqdm import tqdm
from jax import jit, vmap, grad, random, lax
import jax.numpy as jnp
from jax.experimental import host_callback


def progress_bar_scan(num_samples, message=None):
    "Progress bar for a JAX scan"
    if message is None:
        message = f"Running for {num_samples:,} iterations"
    tqdm_bars = tqdm(range(num_samples))
    tqdm_bars.set_description(message)

    if num_samples > 20:
        print_rate = int(num_samples / 20)
    else:
        print_rate = 1  # if you run the sampler for less than 20 iterations

    print_rate = 1
    remainder = num_samples % print_rate

    def _define_tqdm(arg, transform):
        tqdm_bars = tqdm(range(num_samples))
        tqdm_bars.set_description(message)

    def _update_tqdm(arg, transform):
        tqdm_bars.update(arg)

    def _update_progress_bar(iter_num):
        "Updates tqdm progress bar of a JAX scan or loop"
        _ = lax.cond(
            iter_num == 0,
            lambda _: host_callback.id_tap(_define_tqdm, None, result=iter_num),
            lambda _: iter_num,
            operand=None,
        )

        _ = lax.cond(
            # update tqdm every multiple of `print_rate` except at the end
            (iter_num % print_rate == 0) & (iter_num != num_samples - remainder),
            lambda _: host_callback.id_tap(_update_tqdm, print_rate, result=iter_num),
            lambda _: iter_num,
            operand=None,
        )

        _ = lax.cond(
            # update tqdm by `remainder`
            iter_num == num_samples - remainder,
            lambda _: host_callback.id_tap(_update_tqdm, remainder, result=iter_num),
            lambda _: iter_num,
            operand=None,
        )

    def _close_tqdm(arg, transform):
        tqdm_bars.close()

    def close_tqdm(result, iter_num):
        return lax.cond(
            iter_num == num_samples - 1,
            lambda _: host_callback.id_tap(_close_tqdm, None, result=result),
            lambda _: result,
            operand=None,
        )

    def _progress_bar_scan(func):
        """Decorator that adds a progress bar to `body_fun` used in `lax.scan`.
        Note that `body_fun` must either be looping over `np.arange(num_samples)`,
        or be looping over a tuple who's first element is `np.arange(num_samples)`
        This means that `iter_num` is the current iteration number
        """

        def wrapper_progress_bar(iter_num, state):
            _update_progress_bar(iter_num)
            result = func(iter_num, state)
            return close_tqdm(result, iter_num)

        return wrapper_progress_bar

    return _progress_bar_scan


# @progress_bar_scan(10000)
# def loop_body(i, s):
#   s += 1
#   return s



import jax
import jax.numpy as jnp


def initialize_rng_keys(n_chains, seed=42):
    """
    Initialize the random number generator keys for the sampler.

    Args:
        n_chains (int): Number of chains for the local sampler.
        seed (int): Seed for the random number generator.


    Returns:
        rng_keys_init (Device Array): RNG keys for sampling initial position from prior.
        rng_keys_mcmc (Device Array): RNG keys for the local sampler.
        rng_keys_nf (Device Array): RNG keys for the normalizing flow global sampler.
        init_rng_keys_nf (Device Array): RNG keys for initializing wieght of the normalizing flow model.
    """
    rng_key = jax.random.PRNGKey(seed)
    rng_key_init, rng_key_mcmc, rng_key_nf = jax.random.split(rng_key, 3)

    rng_keys_mcmc = jax.random.split(rng_key_mcmc, n_chains)  # (nchains,)
    rng_keys_nf, init_rng_keys_nf = jax.random.split(rng_key_nf, 2)

    return rng_key_init, rng_keys_mcmc, rng_keys_nf, init_rng_keys_nf



# from https://github.com/jwalton3141/jwalton3141.github.io/blob/master/assets/posts/ESS/rwmh.py
# ref Gelman, Andrew, J. B. Carlin, Hal S. Stern, David B. Dunson, Aki Vehtari, and Donald B. Rubin. 2013. Bayesian Data Analysis. Third Edition. London: Chapman & Hall / CRC Press.

import jax.numpy as jnp


def ESS(x):
    """
    Compute the effective sample size of estimand of interest.
    Vectorised implementation.
    x : m_chaines, n_iter, dim

    Output:
    list of ints of length dim: ESS for each dims
    """
    if x.shape < (2,):
        raise ValueError(
            "Calculation of effective sample size"
            "requires multiple chains of the same length."
        )
    try:
        m_chains, n_iter = x.shape  # 1d variable
    except ValueError:
        return [ESS(y.T) for y in x.T]  # apply sequentially to each dimension

    def variogram(t):
        return ((x[:, t:] - x[:, : (n_iter - t)]) ** 2).sum() / (
            m_chains * (n_iter - t)
        )

    post_var = gelman_rubin(x)
    assert post_var > 0

    t = 1
    rho = jnp.ones(n_iter)
    negative_autocorr = False

    # Iterate until the sum of consecutive estimates of autocorrelation is negative
    while not negative_autocorr and (t < n_iter):
        rho = rho.at[t].set(1 - variogram(t) / (2 * post_var))

        if not t % 2:
            negative_autocorr = sum(rho[t - 1 : t + 1]) < 0

        t += 1

    return int(m_chains * n_iter / (1 + 2 * rho[1:t].sum()))


def gelman_rubin(x):
    """
    Estimate the marginal posterior variance. Vectorised implementation.
    x : m_chaines, n_iter
    """
    m_chains, n_iter = x.shape

    # Calculate between-chain variance
    B_over_n = ((jnp.mean(x, axis=1) - jnp.mean(x)) ** 2).sum() / (m_chains - 1)

    # Calculate within-chain variances
    W = ((x - x.mean(axis=1, keepdims=True)) ** 2).sum() / (m_chains * (n_iter - 1))

    # (over) estimate of variance
    s2 = W * (n_iter - 1) / n_iter + B_over_n

    return s2



import warnings
from functools import wraps
from typing import Any, List, Dict, Callable, Iterable, Tuple, NamedTuple, Union

Array = Any
PyTree = Union[Array, Iterable[Array], Dict[Any, Array], NamedTuple]
SampleStats = Dict[str, Array]
Extras = PyTree

import jax
import jax.numpy as jnp
import numpy as np
from jax import lax
from jax._src import dtypes
from jax._src.util import safe_zip
from jax.custom_batching import custom_vmap
from jax.experimental import host_callback
from jax.tree_util import tree_flatten, tree_unflatten


def wrap_python_log_prob_fn(python_log_prob_fn: Callable[..., Array]):
    @custom_vmap
    @wraps(python_log_prob_fn)
    def log_prob_fn(params: Array) -> Array:
        dtype = _tree_dtype(params)
        return host_callback.call(
            python_log_prob_fn,
            params,
            result_shape=jax.ShapeDtypeStruct((), dtype),
        )

    @log_prob_fn.def_vmap
    def _(axis_size: int, in_batched: List[bool], params: Array) -> Tuple[Array, bool]:
        del axis_size, in_batched

        if _arraylike(params):
            flat_params = params
            eval_one = python_log_prob_fn
        else:
            flat_params, unravel = ravel_ensemble(params)
            eval_one = lambda x: python_log_prob_fn(unravel(x))

        result_shape = jax.ShapeDtypeStruct((flat_params.shape[0],), flat_params.dtype)

        result = host_callback.call(
            lambda y: np.stack([eval_one(x) for x in y]),
            flat_params,
            result_shape=result_shape,
        )
        return (
            result,
            True,
        )

    return log_prob_fn


def _tree_dtype(tree: PyTree) -> Any:
    leaves, _ = tree_flatten(tree)
    from_dtypes = [dtypes.dtype(l) for l in leaves]
    return dtypes.result_type(*from_dtypes)


def _arraylike(x: Array) -> bool:
    return (
        isinstance(x, np.ndarray)
        or isinstance(x, jnp.ndarray)
        or hasattr(x, "__jax_array__")
        or np.isscalar(x)
    )


UnravelFn = Callable[[Array], PyTree]

zip = safe_zip


def ravel_ensemble(coords: PyTree) -> Tuple[Array, UnravelFn]:
    leaves, treedef = tree_flatten(coords)
    flat, unravel_inner = _ravel_inner(leaves)
    unravel_one = lambda flat: tree_unflatten(treedef, unravel_inner(flat))
    return flat, unravel_one


def _ravel_inner(lst: List[Array]) -> Tuple[Array, UnravelFn]:
    if not lst:
        return jnp.array([], jnp.float32), lambda _: []
    from_dtypes = [dtypes.dtype(l) for l in lst]
    to_dtype = dtypes.result_type(*from_dtypes)
    shapes = [jnp.shape(x)[1:] for x in lst]
    indices = np.cumsum([int(np.prod(s)) for s in shapes])

    if all(dt == to_dtype for dt in from_dtypes):
        del from_dtypes, to_dtype

        def unravel(arr: Array) -> PyTree:
            chunks = jnp.split(arr, indices[:-1])
            return [chunk.reshape(shape) for chunk, shape in zip(chunks, shapes)]

        ravel = lambda arg: jnp.concatenate([jnp.ravel(e) for e in arg])
        raveled = jax.vmap(ravel)(lst)
        return raveled, unravel

    else:

        def unravel(arr: Array) -> PyTree:
            arr_dtype = dtypes.dtype(arr)
            if arr_dtype != to_dtype:
                raise TypeError(
                    f"unravel function given array of dtype {arr_dtype}, "
                    f"but expected dtype {to_dtype}"
                )
            chunks = jnp.split(arr, indices[:-1])
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                return [
                    lax.convert_element_type(chunk.reshape(shape), dtype)
                    for chunk, shape, dtype in zip(chunks, shapes, from_dtypes)
                ]

        ravel = lambda arg: jnp.concatenate(
            [jnp.ravel(lax.convert_element_type(e, to_dtype)) for e in arg]
        )
        raveled = jax.vmap(ravel)(lst)
        return raveled, unravel



from logging import lastResort
from typing import Callable, Tuple
import jax
import jax.numpy as jnp
import numpy as np
from flowMC.nfmodel.utils import sample_nf, make_training_loop
from flowMC.sampler.NF_proposal import make_nf_metropolis_sampler
from flax.training import train_state  # Useful dataclass to keep train state
import flax
import optax


class Sampler():
    """
    Sampler class that host configuration parameters, NF model, and local sampler

    Args:
        n_dim (int): Dimension of the problem.
        rng_key_set (Tuple): Tuple of random number generator keys.
        local_sampler (Callable): Local sampler maker
        sampler_params (dict): Parameters for the local sampler.
        likelihood (Callable): Likelihood function.
        nf_model (Callable): Normalizing flow model.
        n_loop_training (int, optional): Number of training loops. Defaults to 2.
        n_loop_production (int, optional): Number of production loops. Defaults to 2.
        n_local_steps (int, optional): Number of local steps per loop. Defaults to 5.
        n_global_steps (int, optional): Number of global steps per loop. Defaults to 5.
        n_chains (int, optional): Number of chains. Defaults to 5.
        n_epochs (int, optional): Number of epochs per training loop. Defaults to 5.
        learning_rate (float, optional): Learning rate for the NF model. Defaults to 0.01.
        max_samples (int, optional): Maximum number of samples fed to training the NF model. Defaults to 10000.
        momentum (float, optional): Momentum for the NF model. Defaults to 0.9.
        batch_size (int, optional): Batch size for the NF model. Defaults to 10.
        use_global (bool, optional): Whether to use global sampler. Defaults to True.
        logging (bool, optional): Whether to log the training process. Defaults to True.
        nf_variable (None, optional): Mean and variance variables for the NF model. Defaults to None.
        keep_quantile (float, optional): Quantile of chains to keep when training the normalizing flow model. Defaults to 0.5.
        local_autotune (None, optional): Auto-tune function for the local sampler. Defaults to None.

    """

    def __init__(
        self,
        n_dim: int,
        rng_key_set: Tuple,
        local_sampler: Callable,
        sampler_params: dict,
        likelihood: Callable,
        nf_model: Callable,
        n_loop_training: int = 3,
        n_loop_production: int = 3,
        n_local_steps: int = 50,
        n_global_steps: int = 50,
        n_chains: int = 20,
        n_epochs: int = 30,
        learning_rate: float = 0.01,
        max_samples: int = 10000,
        momentum: float = 0.9,
        batch_size: int = 10000,
        use_global: bool = True,
        logging: bool = True,
        nf_variable=None,
        keep_quantile=0,
        local_autotune=None,
        train_thinning = 1,
    ):
        rng_key_init, rng_keys_mcmc, rng_keys_nf, init_rng_keys_nf = rng_key_set

        self.likelihood = likelihood
        self.likelihood_vec = jax.jit(jax.vmap(self.likelihood))
        self.sampler_params = sampler_params
        self.local_sampler = local_sampler(likelihood)
        self.local_autotune = local_autotune

        self.rng_keys_nf = rng_keys_nf
        self.rng_keys_mcmc = rng_keys_mcmc
        self.n_dim = n_dim
        self.n_loop_training = n_loop_training
        self.n_loop_production = n_loop_production
        self.n_local_steps = n_local_steps
        self.n_global_steps = n_global_steps
        self.n_chains = n_chains
        self.n_epochs = n_epochs
        self.learning_rate = learning_rate
        self.max_samples = max_samples
        self.momentum = momentum
        self.batch_size = batch_size
        self.use_global = use_global
        self.logging = logging


        self.nf_model = nf_model
        model_init = nf_model.init(init_rng_keys_nf, jnp.ones((1, self.n_dim)))
        params = model_init["params"]
        self.variables = model_init["variables"]
        if nf_variable is not None:
            self.variables = self.variables

        self.keep_quantile = keep_quantile
        self.train_thinning = train_thinning
        self.nf_training_loop, train_epoch, train_step = make_training_loop(
            self.nf_model
        )
        self.global_sampler = make_nf_metropolis_sampler(self.nf_model)

        tx = optax.adam(self.learning_rate, self.momentum)
        self.state = train_state.TrainState.create(
            apply_fn=nf_model.apply, params=params, tx=tx
        )

        training = {}
        training["chains"] = jnp.empty((self.n_chains, 0, self.n_dim))
        training["log_prob"] = jnp.empty((self.n_chains, 0))
        training["local_accs"] = jnp.empty((self.n_chains, 0))
        training["global_accs"] = jnp.empty((self.n_chains, 0))
        training["loss_vals"] = jnp.empty((0, self.n_epochs))

        production = {}
        production["chains"] = jnp.empty((self.n_chains, 0, self.n_dim))
        production["log_prob"] = jnp.empty((self.n_chains, 0))
        production["local_accs"] = jnp.empty((self.n_chains, 0))
        production["global_accs"] = jnp.empty((self.n_chains, 0))

        self.summary = {}
        self.summary['training'] = training
        self.summary['production'] = production

    def sample(self, initial_position):
        """
        Sample from the posterior using the local sampler.

        Args:
            initial_position (Device Array): Initial position.

        Returns:
            chains (Device Array): Samples from the posterior.
            nf_samples (Device Array): (n_nf_samples, n_dim)
            local_accs (Device Array): (n_chains, n_local_steps * n_loop)
            global_accs (Device Array): (n_chains, n_global_steps * n_loop)
            loss_vals (Device Array): (n_epoch * n_loop,)
        """

        # Note that auto-tune function needs to have the same number of steps
        # as the actual sampling loop to avoid recompilation.

        self.local_sampler_tuning(self.n_local_steps, initial_position)
        last_step = initial_position
        if self.use_global == True:
            last_step = self.global_sampler_tuning(last_step)

        last_step = self.production_run(last_step)

    def sampling_loop(self, initial_position: jnp.array, training=False) -> jnp.array:
        """
        One sampling loop that iterate through the local sampler and potentially the global sampler.
        If training is set to True, the loop will also train the normalizing flow model.

        Args:
            initial_position (jnp.array): Initial position. Shape (n_chains, n_dim)
            training (bool, optional): Whether to train the normalizing flow model. Defaults to False.

        Returns:
            chains (jnp.array): Samples from the posterior. Shape (n_chains, n_local_steps + n_global_steps, n_dim)

        """

        if training == True:
            summary_mode = 'training'
        else:
            summary_mode = 'production'

        self.rng_keys_mcmc, positions, log_prob, local_acceptance, _ = self.local_sampler(
            self.rng_keys_mcmc, self.n_local_steps, initial_position, self.sampler_params
        )

        self.summary[summary_mode]['chains'] = jnp.append(
            self.summary[summary_mode]['chains'], positions, axis=1
        )
        self.summary[summary_mode]['log_prob'] = jnp.append(
            self.summary[summary_mode]['log_prob'], log_prob, axis=1
        )
        self.summary[summary_mode]['local_accs'] = jnp.append(
            self.summary[summary_mode]['local_accs'], local_acceptance, axis=1
        )

        if self.use_global == True:
            if training == True:
                positions = self.summary['training']['chains'][:,::self.train_thinning]
                log_prob_output = self.summary['training']['log_prob'][:,::self.train_thinning]

                if self.keep_quantile > 0:
                    max_log_prob = jnp.max(log_prob_output, axis=1)
                    cut = jnp.quantile(max_log_prob, self.keep_quantile)
                    cut_chains = positions[max_log_prob > cut]
                else:
                    cut_chains = positions
                chain_size = cut_chains.shape[0] * cut_chains.shape[1]
                if chain_size > self.max_samples:
                    flat_chain = cut_chains[
                        :, -int(self.max_samples / self.n_chains):
                    ].reshape(-1, self.n_dim)
                else:
                    flat_chain = cut_chains.reshape(-1, self.n_dim)

                if flat_chain.shape[0] < self.max_samples:
                    # This is to pad the training data to avoid recompilation.
                    flat_chain = jnp.repeat(flat_chain, (self.max_samples // flat_chain.shape[0])+1, axis=0)
                    flat_chain = flat_chain[:self.max_samples]

                variables = self.variables.unfreeze()
                variables["base_mean"] = jnp.mean(flat_chain, axis=0)
                variables["base_cov"] = jnp.cov(flat_chain.T)
                self.variables = flax.core.freeze(variables)

                self.rng_keys_nf, self.state, loss_values = self.nf_training_loop(
                    self.rng_keys_nf,
                    self.state,
                    self.variables,
                    flat_chain,
                    self.n_epochs,
                    self.batch_size,
                )
                self.summary['training']['loss_vals'] = jnp.append(
                    self.summary['training']['loss_vals'], loss_values.reshape(1, -1), axis=0
                )

            (
                self.rng_keys_nf,
                nf_chain,
                log_prob,
                log_prob_nf,
                global_acceptance,
            ) = self.global_sampler(
                self.rng_keys_nf,
                self.n_global_steps,
                self.state.params,
                self.variables,
                self.likelihood_vec,
                positions[:, -1],
            )

            self.summary[summary_mode]['chains'] = jnp.append(
                self.summary[summary_mode]['chains'], nf_chain, axis=1
            )
            self.summary[summary_mode]['log_prob'] = jnp.append(
                self.summary[summary_mode]['log_prob'], log_prob, axis=1
            )
            self.summary[summary_mode]['global_accs'] = jnp.append(
                self.summary[summary_mode]['global_accs'], global_acceptance, axis=1
            )

        last_step = self.summary[summary_mode]['chains'][:, -1]

        return last_step

    def local_sampler_tuning(self, n_steps: int, initial_position: jnp.array, max_iter: int = 10):
        """
        Tuning the local sampler. This runs a number of iterations of the local sampler,
        and then uses the acceptance rate to adjust the local sampler parameters.
        Since this is mostly for a fast adaptation, we do not carry the sample state forward.
        Instead, we only adapt the sampler parameters using the initial position.

        Args:
            n_steps (int): Number of steps to run the local sampler.
            initial_position (Device Array): Initial position for the local sampler.
            max_iter (int): Number of iterations to run the local sampler.
        """
        if self.local_autotune is not None:
            print("Autotune found, start tuning sampler_params")
            self.sampler_params, self.local_sampler = self.local_autotune(
                self.local_sampler, self.rng_keys_mcmc, n_steps, initial_position, self.sampler_params, max_iter)
        else:
            print("No autotune found, use input sampler_params")

    def global_sampler_tuning(self, initial_position: jnp.ndarray) -> jnp.array:
        """
        Tuning the global sampler. This runs both the local sampler and the global sampler,
        and train the normalizing flow on the run.
        To adapt the normalizing flow, we need to keep certain amount of the data generated during the sampling.
        The data is stored in the summary dictionary.

        Args:
            initial_position (Device Array): Initial position for the sampler, shape (n_chains, n_dim)

        """
        print("Training normalizing flow")
        last_step = initial_position
        for _ in range(self.n_loop_training):
            last_step = self.sampling_loop(last_step, training=True)
        return last_step

    def production_run(self, initial_position: jnp.ndarray):
        """
        Sampling procedure that produce the final set of samples.
        The main difference between this and the global tuning step is
        we do not train the normalizing flow in order to main detail balance.
        The data is stored in the summary dictionary.
        
        """
        last_step = initial_position
        for _ in range(self.n_loop_production):
            self.sampling_loop(last_step)

    def get_sampler_state(self, training: bool=False) -> dict:
        """
        Get the sampler state. There are two sets of sampler outputs one can get,
        the training set and the production set.
        The training set is produced during the global tuning step, and the production set
        is produced during the production run.
        Only the training set contains information about the loss function.
        Only the production set should be used to represent the final set of samples.

        Args:
            training (bool): Whether to get the training set sampler state. Defaults to False.
        
        """
        if training == True:
            return self.summary['training']
        else:
            return self.summary['production']

    def sample_flow(self, n_samples: int) -> jnp.ndarray:
        """

        Sample from the normalizing flow.

        Args:
            n_samples (int): Number of samples to generate.

        Returns:
            Device Array: Samples generated using the normalizing flow.
        """

        nf_samples = sample_nf(
            self.nf_model,
            self.state.params,
            self.rng_keys_nf,
            n_samples,
            self.variables,
        )
        return nf_samples

    def reset(self):
        """
        Reset the sampler state.

        """
        training = {}
        training["chains"] = jnp.empty((self.n_chains, 0, self.n_dim))
        training["log_prob"] = jnp.empty((self.n_chains, 0))
        training["local_accs"] = jnp.empty((self.n_chains, 0))
        training["global_accs"] = jnp.empty((self.n_chains, 0))
        training["loss_vals"] = jnp.empty((0, self.n_epochs))

        production = {}
        production["chains"] = jnp.empty((self.n_chains, 0, self.n_dim))
        production["log_prob"] = jnp.empty((self.n_chains, 0))
        production["local_accs"] = jnp.empty((self.n_chains, 0))
        production["global_accs"] = jnp.empty((self.n_chains, 0))

        self.summary = {}
        self.summary['training'] = training
        self.summary['production'] = production



from typing import Callable
import jax
import jax.numpy as jnp
from jax.scipy.stats import multivariate_normal
from tqdm import tqdm


def make_mala_kernel(logpdf: Callable):
    """
    Making MALA kernel for a single step

    Args:
        logpdf (Callable): log-probability density function


    Returns:
        mala_kernel (Callable): MALA kernel for a single step

    """

    def body(carry, this_key):
        this_position, dt = carry
        dt2 = dt*dt
        this_log_prob, this_d_log = jax.value_and_grad(logpdf)(this_position)
        proposal = this_position + jnp.dot(dt2, this_d_log) / 2
        proposal += jnp.dot(dt, jax.random.normal(this_key, shape=this_position.shape))
        return (proposal,dt), (proposal, this_log_prob, this_d_log)

    def mala_kernel(rng_key, position, log_prob, dt = 1e-1):

        """
        Metropolis-adjusted Langevin algorithm kernel.
        This function make a proposal and accept/reject it.

        Args:
            rng_key (n_chains, 2): random key
            logpdf (function) : log-density function
            d_logpdf (function): gradient of log-density function
            position (n_chains, n_dim): current position
            log_prob (n_chains, ): log-probability of the current position
            dt (float): step size of the MALA step

        Returns:
            position (n_chains, n_dim): the new poisiton of the chain
            log_prob (n_chains, ): the log-probability of the new position
            do_accept (n_chains, ): whether to accept the new position

        """
        key1, key2 = jax.random.split(rng_key)

        dt2 = dt*dt

        _, (proposal, logprob, d_logprob) = jax.lax.scan(
            body, (position, dt), jnp.array([key1, key1])
        )

        ratio = logprob[1] - logprob[0]
        ratio -= multivariate_normal.logpdf(
            proposal[0], position + jnp.dot(dt2, d_logprob[0]) / 2, dt2
        )
        ratio += multivariate_normal.logpdf(
            position, proposal[0] + jnp.dot(dt2, d_logprob[1]) / 2, dt2
        )

        log_uniform = jnp.log(jax.random.uniform(key2))
        do_accept = log_uniform < ratio

        position = jnp.where(do_accept, proposal[0], position)
        log_prob = jnp.where(do_accept, logprob[1], logprob[0])
        return position, log_prob, do_accept

    return mala_kernel


def make_mala_update(logpdf):
    """

    Making MALA update function for multiple steps

    Args:
        logpdf (Callable): log-probability density function
        d_logpdf (Callable): gradient of log-probability density function
        dt (float): step size of the MALA step
        M (jnp.array): mass matrix. Currently we only support diagonal mass matrix

    Returns:
        mala_update (Callable): MALA update function for multiple steps

    """
    mala_kernel = make_mala_kernel(logpdf)

    def mala_update(i, state):
        key, positions, log_prob, acceptance, dt = state
        _, key = jax.random.split(key)
        new_position, new_log_prob, do_accept = mala_kernel(
            key, positions[i - 1], log_prob[i - 1], dt
        )
        positions = positions.at[i].set(new_position)
        log_prob = log_prob.at[i].set(new_log_prob)
        acceptance = acceptance.at[i].set(do_accept)
        return (key, positions, log_prob, acceptance, dt)

    # Apperantly jitting after vmap will make compilation much slower.
    # Output the kernel, logpdf, and dlogpdf for warmup jitting.
    # Apperantly passing in a warmed up function will still trigger recompilation.
    # so the warmup need to be done with the output function

    return mala_update, logpdf


def make_mala_sampler(logpdf: Callable, jit: bool=False):
    mala_update, lp = make_mala_update(logpdf)
    # Somehow if I define the function inside the other function,
    # I think it doesn't use the cache and recompile everytime.
    if jit:
        mala_update = jax.jit(mala_update)
        lp = jax.jit(lp)

    mala_update = jax.vmap(mala_update, in_axes=(None, (0, 0, 0, 0, None)), out_axes=(0, 0, 0, 0, None))
    lp = jax.vmap(lp)

    def mala_sampler(rng_key, n_steps, initial_position, sampler_params={'dt':1e-1}):

        """
        Metropolis-adjusted Langevin algorithm sampler.
        This function do n step with the MALA kernel.

        Args:
            rng_key (n_chains, 2): random key for the sampler
            n_steps (int): number of local steps
            logpdf (function): log-density function
            d_logpdf (function): gradient of log-density function
            initial_position (n_chains, n_dim): initial position of the chain
            dt (float): step size of the MALA step

        Returns:
            rng_key (n_chains, 2): random key for the sampler after the sampling
            all_positions (n_chains, n_steps, n_dim): all the positions of the chain
            log_probs (n_chains, ): log probability at the end of the chain
            acceptance: acceptance rate of the chain
        """

        logp = lp(initial_position)
        n_chains = rng_key.shape[0]
        acceptance = jnp.zeros(
            (
                n_chains,
                n_steps,
            )
        )
        all_positions = (
            jnp.zeros(
                (
                    n_chains,
                    n_steps,
                )
                + initial_position.shape[-1:]
            )
            + initial_position[:, None]
        )
        all_logp = (
            jnp.zeros(
                (
                    n_chains,
                    n_steps,
                )
            )
            + logp[:, None]
        )
        state = (rng_key, all_positions, all_logp, acceptance, sampler_params['dt'])
        for i in tqdm(
            range(1, n_steps), desc="Sampling Locally", miniters=int(n_steps / 10)
        ):
            state = mala_update(i, state)
        return state

    return mala_sampler
    
from tqdm import tqdm
from functools import partialmethod

def mala_sampler_autotune(mala_sampler, rng_key, n_steps, initial_position, sampler_params, max_iter = 30):
    tqdm.__init__ = partialmethod(tqdm.__init__, disable=True)

    counter = 0
    dt = sampler_params['dt']
    rng_keys_mcmc, positions, log_prob, local_acceptance, _ = mala_sampler(rng_key, n_steps, initial_position, {'dt':dt})
    acceptance_rate = jnp.mean(local_acceptance)
    while (acceptance_rate < 0.3) or (acceptance_rate > 0.5):
        if counter > max_iter:
            print("Maximal number of iterations reached. Existing tuning with current parameters.")
            break
        if acceptance_rate < 0.3:
            dt *= 0.8
        elif acceptance_rate > 0.5:
            dt *= 1.25
        counter += 1
        rng_keys_mcmc, positions, log_prob, local_acceptance, _ = mala_sampler(rng_keys_mcmc, n_steps, initial_position, {'dt':dt})
        acceptance_rate = jnp.mean(local_acceptance)
    tqdm.__init__ = partialmethod(tqdm.__init__, disable=False)
    return {"dt": dt}, mala_sampler


################### Scan API ##############################

# def make_mala_kernel(logpdf, d_logpdf, dt):
#     def mala_kernel(carry, data):
#         rng_key, position, log_prob, do_accept = carry
#         rng_key, key1, key2 = jax.random.split(rng_key,3)
#         proposal = position + dt * d_logpdf(position)
#         proposal += dt * jnp.sqrt(2/dt) * jax.random.normal(key1, shape=position.shape)
#         ratio = logpdf(proposal) - logpdf(position)
#         ratio -= ((position - proposal - dt * d_logpdf(proposal)) ** 2 / (4 * dt)).sum()
#         ratio += ((proposal - position - dt * d_logpdf(position)) ** 2 / (4 * dt)).sum()
#         proposal_log_prob = logpdf(proposal)

#         log_uniform = jnp.log(jax.random.uniform(key2))
#         do_accept = log_uniform < ratio

#         position = jax.lax.cond(do_accept, lambda: proposal, lambda: position)
#         log_prob = jax.lax.cond(do_accept, lambda: proposal_log_prob, lambda: log_prob)
#         return (rng_key, position, log_prob, do_accept), (position, log_prob, do_accept)
#     return mala_kernel, logpdf, d_logpdf

# def make_mala_update(logpdf, d_logpdf, dt):
#     mala_kernel, logpdf, d_logpdf = make_mala_kernel(logpdf, d_logpdf, dt)
#     def mala_update(rng_key, position, logp, n_steps=100):
#         carry = (rng_key, position, logp, False)
#         y = jax.lax.scan(mala_kernel, carry, jax.random.split(rng_key,n_steps))
#         return y
#     mala_update = jax.vmap(mala_update, in_axes=(0,0,0,None))
#     mala_kernel_vec = jax.vmap(mala_kernel, in_axes=((0,0,0,0),None))
#     logpdf = jax.vmap(logpdf)
#     d_logpdf = jax.vmap(d_logpdf)
#     return mala_update, mala_kernel_vec, logpdf, d_logpdf



from typing import Tuple
import jax
import jax.numpy as jnp
from jax import random, jit, vmap
from tqdm import tqdm

n_sample_max = 100000


def nf_metropolis_kernel(
    rng_key: jax.random.PRNGKey,
    proposal_position: jnp.ndarray,
    initial_position: jnp.ndarray,
    proposal_pdf: jnp.ndarray,
    proposal_nf_pdf: jnp.ndarray,
    initial_pdf: jnp.ndarray,
    initial_nf_pdf: jnp.ndarray,
):

    """
    A one-step global sampling kernel for a given normalizing flow model.

    Args:
        rng_key: Jax PRNGKey.
        proposal_position: Proposed positions, shape (Ndim).
        initial_position: Initial positions, shape (Ndim).
        proposal_pdf: Pdf value evaluate using the target function at the proposal position, shape (Ndim).
        proposal_nf_pdf: Pdf value evaluate using the normalizing flow model at the proposal position, shape (Ndim).
        initial_pdf: Pdf value evaluate using the target function at the initial position, shape (Ndim).
        initial_nf_pdf: Pdf value evaluate using the normalizing flow model at the initial position, shape (Ndim).

    Returns:
        position: New positions, shape (Ndim).
        log_prob: Pdf value evaluate using the target function at the new position, shape (Ndim).
        log_prob_nf: Pdf value evaluate using the normalizing flow model at the new position, shape (Ndim).
        do_accept: Acceptance boolean, shape (Ndim).
    """

    rng_key, subkeys = random.split(rng_key, 2)
    ratio = (proposal_pdf - initial_pdf) - (proposal_nf_pdf - initial_nf_pdf)
    u = jnp.log(jax.random.uniform(subkeys, ratio.shape))
    do_accept = u < ratio
    position = jnp.where(do_accept, proposal_position, initial_position)
    log_prob = jnp.where(do_accept, proposal_pdf, initial_pdf)
    log_prob_nf = jnp.where(do_accept, proposal_nf_pdf, initial_nf_pdf)
    return position, log_prob, log_prob_nf, do_accept


nf_metropolis_kernel = vmap(nf_metropolis_kernel)


@jax.jit
def nf_metropolis_update(i: int, state: Tuple):

    """
    A multistep global sampling kernel for a given normalizing flow model.

    Args:
        i: Number of iteration.
        state: A tuple containing the current state of the sampler.

    """

    (
        key,
        positions,
        proposal,
        log_prob,
        log_prob_nf,
        log_prob_proposal,
        log_prob_nf_proposal,
        acceptance,
    ) = state
    key, *sub_key = jax.random.split(key, positions.shape[1] + 1)
    sub_key = jnp.array(sub_key)
    new_position, new_log_prob, new_log_prob_nf, do_accept = nf_metropolis_kernel(
        sub_key,
        proposal[i],
        positions[i - 1],
        log_prob_proposal[i],
        log_prob_nf_proposal[i],
        log_prob[i - 1],
        log_prob_nf[i - 1],
    )
    positions = positions.at[i].set(new_position)
    log_prob = log_prob.at[i].set(new_log_prob)
    log_prob_nf = log_prob_nf.at[i].set(new_log_prob_nf)

    acceptance = acceptance.at[i].set(do_accept)

    return (
        key,
        positions,
        proposal,
        log_prob,
        log_prob_nf,
        log_prob_proposal,
        log_prob_nf_proposal,
        acceptance,
    )


def make_nf_metropolis_sampler(nf_model):

    """
    A function make the normalizing flow sampler for a given normalizing flow model.

    Args:
        nf_model: A normalizing flow model.
        
    Returns:
        nf_metropolis_sampler: A normalizing flow sampler that has the following signature:

            Args:
                rng_key: Jax PRNGKey.
                n_steps: Number of steps.
                nf_param: Normalizing flow parameters.
                nf_variables: Normalizing flow variables.
                target_pdf: Target pdf.
                initial_position: Initial position.

            Returns:
                rng_key: Current state of random key
                all_positions: All positions of the chain
                all_logp: Log probability of the target function at all positions
                all_logp_nf: Log probability of the normalizing flow model at all positions
                acceptance: Acceptance boolean

    """

    @jax.jit
    def eval_nf_logprob(position, nf_params, nf_variables):
        return nf_model.apply(
            {"params": nf_params, "variables": nf_variables},
            position,
            method=nf_model.log_prob,
        )

    def sample_nf(rng_key, n_samples, nf_params, nf_variables):
        return nf_model.apply(
            {"params": nf_params, "variables": nf_variables},
            rng_key,
            n_samples,
            method=nf_model.sample,
        )

    sample_nf = jax.jit(sample_nf, static_argnums=(1))

    def nf_metropolis_sampler(
        rng_key, n_steps, nf_param, nf_variables, target_pdf, initial_position
    ):
        rng_key, *subkeys = random.split(rng_key, 3)

        total_sample = initial_position.shape[0] * n_steps

        log_pdf_nf_initial = eval_nf_logprob(
            initial_position, nf_param, nf_variables
        )
        log_pdf_initial = target_pdf(initial_position)

        if total_sample > n_sample_max:
            proposal_position = jnp.zeros((total_sample, initial_position.shape[-1]))
            log_pdf_nf_proposal = jnp.zeros((total_sample,))
            log_pdf_proposal = jnp.zeros((total_sample,))
            local_key, subkey = random.split(subkeys[0], 2)
            for i in tqdm(
                range(total_sample // n_sample_max),
                desc="Sampling Globally",
                miniters=(total_sample // n_sample_max) // 10,
            ):
                local_samples = sample_nf(subkey, n_sample_max, nf_param, nf_variables)
                proposal_position = proposal_position.at[
                    i * n_sample_max : (i + 1) * n_sample_max
                ].set(local_samples)
                log_pdf_nf_proposal = log_pdf_nf_proposal.at[
                    i * n_sample_max : (i + 1) * n_sample_max
                ].set(eval_nf_logprob(local_samples, nf_param, nf_variables))
                log_pdf_proposal = log_pdf_proposal.at[
                    i * n_sample_max : (i + 1) * n_sample_max
                ].set(target_pdf(local_samples))
                local_key, subkey = random.split(local_key, 2)

        else:
            proposal_position = sample_nf(
                subkeys[0], total_sample, nf_param, nf_variables
            )
            log_pdf_nf_proposal = eval_nf_logprob(
                proposal_position, nf_param, nf_variables
            )
            log_pdf_proposal = target_pdf(proposal_position)

        proposal_position = proposal_position.reshape(
            n_steps, initial_position.shape[0], initial_position.shape[1]
        )
        log_pdf_nf_proposal = log_pdf_nf_proposal.reshape(
            n_steps, initial_position.shape[0]
        )
        log_pdf_proposal = log_pdf_proposal.reshape(n_steps, initial_position.shape[0])

        all_positions = (
            jnp.zeros((n_steps,) + initial_position.shape) + initial_position
        )
        all_logp = jnp.zeros((n_steps, initial_position.shape[0])) + log_pdf_initial
        all_logp_nf = (
            jnp.zeros((n_steps, initial_position.shape[0])) + log_pdf_nf_initial
        )
        acceptance = jnp.zeros((n_steps, initial_position.shape[0]))

        initial_state = (
            subkeys[1],
            all_positions,
            proposal_position,
            all_logp,
            all_logp_nf,
            log_pdf_proposal,
            log_pdf_nf_proposal,
            acceptance,
        )
        (
            rng_key,
            all_positions,
            proposal_position,
            all_logp,
            all_logp_nf,
            log_pdf_proposal,
            log_pdf_nf_proposal,
            acceptance,
        ) = jax.lax.fori_loop(1, n_steps, nf_metropolis_update, initial_state)
        all_positions = all_positions.swapaxes(0, 1)
        all_logp = all_logp.swapaxes(0, 1)
        all_logp_nf = all_logp_nf.swapaxes(0, 1)
        acceptance = acceptance.swapaxes(0, 1)

        return rng_key, all_positions, all_logp, all_logp_nf, acceptance

    return nf_metropolis_sampler



from flowMC.sampler.MALA import make_mala_kernel, make_mala_update, make_mala_sampler
from flowMC.sampler.MALA import make_mala_sampler
from flowMC.utils.PRNG_keys import initialize_rng_keys
import jax
import jax.numpy as jnp
from jax.scipy.special import logsumexp

def dual_moon_pe(x):
    """
    Term 2 and 3 separate the distribution and smear it along the first and second dimension
    """
    term1 = 0.5 * ((jnp.linalg.norm(x) - 2) / 0.1) ** 2
    term2 = -0.5 * ((x[:1] + jnp.array([-3.0, 3.0])) / 0.8) ** 2
    term3 = -0.5 * ((x[1:2] + jnp.array([-3.0, 3.0])) / 0.6) ** 2
    return -(term1 - logsumexp(term2) - logsumexp(term3))

n_dim = 5
n_chains = 15

rng_key_set = initialize_rng_keys(n_chains, seed=42)

initial_position = jax.random.normal(rng_key_set[0], shape=(n_chains, n_dim)) * 1

def test_mala_kernel():
    kernel = make_mala_kernel(dual_moon_pe)
    update = make_mala_update(kernel)
    sampler = make_mala_sampler(dual_moon_pe,jit=True)

    kernel(rng_key_set[1][0], initial_position[0], dual_moon_pe(initial_position[0]), 1e-1)
    sampler(rng_key_set[1],10,initial_position,{'dt':1e-1})


from flowMC.nfmodel.realNVP import RealNVP
from flowMC.nfmodel.rqSpline import RQSpline
import jax
import jax.numpy as jnp  # JAX NumPy

from flowMC.nfmodel.utils import *
from flax.training import train_state  # Useful dataclass to keep train state
import optax  # Optimizers
import flax

from sklearn.datasets import make_moons
import matplotlib.pyplot as plt



def test_realNVP():

    data = jnp.array(make_moons(n_samples=100, noise=0.05)[0])

    num_epochs = 5
    batch_size = 100
    learning_rate = 0.001
    momentum = 0.9

    key1, rng, init_rng = jax.random.split(jax.random.PRNGKey(0), 3)

    model = RealNVP(2, 2, 16, 1)
    def create_train_state(rng, learning_rate, momentum):
        params = model.init(rng, jnp.ones((1, 2)))["params"]
        tx = optax.adam(learning_rate, momentum)
        return train_state.TrainState.create(apply_fn=model.apply, params=params, tx=tx)

    state = create_train_state(init_rng, learning_rate, momentum)

    variables = model.init(rng, jnp.ones((1, 2)))["variables"]
    variables = variables.unfreeze()
    variables["base_mean"] = jnp.mean(data, axis=0)
    variables["base_cov"] = jnp.cov(data.T)
    variables = flax.core.freeze(variables)

    train_flow, train_epoch, train_step = make_training_loop(model)
    rng, state, loss_values = train_flow(
        rng, state, variables, data, num_epochs, batch_size
    )

    rng_key_nf = jax.random.PRNGKey(124098)
    sample_nf(model, state.params, rng_key_nf, 10000, variables)

def test_rqSpline():


    data = jnp.array(make_moons(n_samples=100, noise=0.05)[0])

    num_epochs = 5
    batch_size = 100
    learning_rate = 0.001
    momentum = 0.9

    key1, rng, init_rng = jax.random.split(jax.random.PRNGKey(0), 3)
    model = RQSpline(2, 4, [16, 16], 8)

    def create_train_state(rng, learning_rate, momentum):
        params = model.init(rng, jnp.ones((1, 2)))["params"]
        tx = optax.adam(learning_rate, momentum)
        return train_state.TrainState.create(apply_fn=model.apply, params=params, tx=tx)

    state = create_train_state(init_rng, learning_rate, momentum)

    variables = model.init(rng, jnp.ones((1, 2)))["variables"]
    variables = variables.unfreeze()
    variables["base_mean"] = jnp.mean(data, axis=0)
    variables["base_cov"] = jnp.cov(data.T)
    variables = flax.core.freeze(variables)

    train_flow, train_epoch, train_step = make_training_loop(model)
    rng, state, loss_values = train_flow(
        rng, state, variables, data, num_epochs, batch_size
    )
    rng_key_nf = jax.random.PRNGKey(124098)
    sample_nf(model, state.params, rng_key_nf, 10000, variables)



from turtle import pos
import jax
import jax.numpy as jnp
from functools import partial

# @partial(jax.jit, static_argnums=(1,))
def rw_metropolis_kernel(rng_key, logpdf, position, log_prob, dt=0.1):

    """Moves the chains by one step using the Random Walk Metropolis algorithm.

    Attributes
    ----------
    rng_key: jax.random.PRNGKey
      Key for the pseudo random number generator.
    logpdf: function
      Returns the log-probability of the model given a position.
    position: jnp.ndarray, shape (n_dims,)
      The starting position.
    log_prob: float
      The log probability at the starting position.
    Returns
    -------
    Tuple
        The next positions of the chains along with their log probability.
    """
    key1, key2 = jax.random.split(rng_key)
    move_proposal = jax.random.normal(key1, shape=position.shape) * dt
    proposal = position + move_proposal
    proposal_log_prob = logpdf(proposal)

    log_uniform = jnp.log(jax.random.uniform(key2))
    do_accept = log_uniform < proposal_log_prob - log_prob

    position = jnp.where(do_accept, proposal, position)
    log_prob = jnp.where(do_accept, proposal_log_prob, log_prob)
    return position, log_prob, do_accept


# @partial(jax.jit, static_argnums=(1, 2))
def rw_metropolis_sampler(rng_key, n_steps, logpdf, initial_position):
    def mh_update_sol2(i, state):
        key, positions, log_prob, acceptance = state
        _, key = jax.random.split(key)
        new_position, new_log_prob, do_accept = rw_metropolis_kernel(
            key, logpdf, positions[i - 1], log_prob[i - 1]
        )
        positions = positions.at[i].set(new_position)
        log_prob = log_prob.at[i].set(new_log_prob)
        acceptance = acceptance.at[i].set(do_accept)
        return (key, positions, log_prob, acceptance)

    logp = logpdf(initial_position)
    acceptance = jnp.zeros((n_steps,))
    all_positions = jnp.zeros((n_steps,) + initial_position.shape) + initial_position
    all_logp = jnp.zeros((n_steps,)) + logp
    initial_state = (rng_key, all_positions, all_logp, acceptance)
    rng_key, all_positions, all_logp, acceptance = jax.lax.fori_loop(
        1, n_steps, mh_update_sol2, initial_state
    )

    return rng_key, all_positions, all_logp, acceptance
